static char *sccsid = "@(#) SupportPacks/ms0c/public/mqsecutl.c, supportpacks, p000, p000-L000330          1.4  99/05/22 23:39:38";
/*---------------------------------------------------------------
 *  Author(s)     : Benjamin Y. Goldshlager
 *
 *  Name: MQSECUTL.C
 *
 *  Description:
 *      This source contain the utilities used by the MQ Channel
 *      Security Exit.
 *
 *
 *  Notes:
 *
 *--------------------------------------------------------------*/
#include "mqsecutl.h"
#include <stdarg.h>

#if MQAT_DEFAULT==MQAT_WINDOWS_NT
#include "mqseccem.h"
#elif MQAT_DEFAULT==MQAT_UNIX
#include <syslog.h>
#else
#error Unsupported platform
#endif

#include <time.h>

#define TIMESTAMP_LEN                           26

#define DEBUG_DEFAULT_MAX_APP_DATA      20
// Maximum allowable number
#define DEBUG_MAX_APP_DATA_VALUE        1000000

#define MQSECCE_INFO_MSG_PREFIX         "MQSECCE(I) "
#define MQSECCE_WARNING_MSG_PREFIX      "MQSECCE(W) "
#define MQSECCE_ERROR_MSG_PREFIX        "MQSECCE(E) "
#define MQSECCE_DEBUG_MSG_PREFIX        "MQSECCE(D) "

#define DEFAULT_IDENT                   "MQSeries Security Channel Exit"

#if MQAT_DEFAULT==MQAT_UNIX
#define DEFAULT_FACILITY                LOG_USER
#define DEFAULT_PRIORITY                LOG_ERR
#endif

enum MsgType_t
{
  DTK_INFO,
  DTK_WARNING,
  DTK_ERROR,
  DTK_DEBUG,
  DTK_DEBUG_NO_PREFIX
};

#define DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN(v1, v2, v3, v4, v5) \
  { \
    dtkError (NULL, "SourceFile <%s> SourceLine <%d> readINIFile(): " \
	"Found invalid value <%s> for parameter <%s> in INI file <%s>\n", \
	v1, v2, v3, v4, v5); \
    return 1; \
  }

#define RETURN_RC_AND_CLOSE_FILE(pF, rc) \
  { \
    fclose (pF); \
    return rc; \
  }

#define ENTRY(param) \
  { param, #param },

typedef struct
{
  MQLONG  lCode;
  char*   szDescription;
} ExitCodeEntry_t;

static ExitCodeEntry_t exitReasonTable[] =
{
  ENTRY(MQXR_INIT)
  ENTRY(MQXR_INIT_SEC)
  ENTRY(MQXR_SEC_MSG)
  ENTRY(MQXR_TERM)
  ENTRY(MQXR_MSG)
  ENTRY(MQXR_XMIT)
  ENTRY(MQXR_RETRY)
  -1,     "Could Not Find Description in exitReasonTable[]"
};

static int exitReasonTableLen = sizeof (exitReasonTable) /
	        sizeof (ExitCodeEntry_t);

static ExitCodeEntry_t exitIdTable[] =
{
  ENTRY(MQXT_CHANNEL_SEC_EXIT)
  ENTRY(MQXT_CHANNEL_MSG_EXIT)
  ENTRY(MQXT_CHANNEL_SEND_EXIT)
  ENTRY(MQXT_CHANNEL_RCV_EXIT)
  ENTRY(MQXT_CHANNEL_MSG_RETRY_EXIT)
  ENTRY(MQXT_CHANNEL_AUTO_DEF_EXIT)
  -1,     "Could Not Find Description in exitIdTable[]"
};

static int exitIdTableLen = sizeof (exitIdTable) /
	      sizeof (ExitCodeEntry_t);

/* Exit Response */
static ExitCodeEntry_t exitResponseTable[] =
{
  ENTRY (MQXCC_OK)
  ENTRY(MQXCC_SUPPRESS_FUNCTION)
  ENTRY(MQXCC_SKIP_FUNCTION)
  ENTRY(MQXCC_SEND_AND_REQUEST_SEC_MSG)
  ENTRY(MQXCC_SEND_SEC_MSG)
  ENTRY(MQXCC_SUPPRESS_EXIT)
  ENTRY(MQXCC_CLOSE_CHANNEL)
  -1,     "Could Not Find Description in exitResponseTable[]"
};

static int exitResponseTableLen = sizeof (exitResponseTable) /
	          sizeof (ExitCodeEntry_t);

static const char* getDescription (MQLONG lCode,
	           ExitCodeEntry_t* table,
	           int iLen);
static void dtkWrite (const PEXITLOCALVARS pLocalVars,
	    enum  MsgType_t msgType,      // Message type, i.e., DTK_INFO, DTK_ERROR...
	    char* lpszMsg);

/*
 * Convert MQCHAR to a string.
 */
char* baseName (char* pszFileName)
{
  int i;
  if (!pszFileName && !pszFileName[0])
    return pszFileName;

  for (i=strlen (pszFileName); i>0; i--)
  {
#if MQAT_DEFAULT==MQAT_UNIX
    if (pszFileName[i] == '/')
#else
    if (pszFileName[i] == '\\')
#endif
      break;
  }

  return &pszFileName[i+1];
}

/*
 * Convert MQCHAR to a string.  This routine assumes that the caller
 * pszString buffer is iLen+1 length to account for the trailing null.
 */
void convertMQCHARToString (char*       pszString,
	      PMQCHAR mqIn,
	      int             iLen)
{
  int i;
  assert (mqIn && pszString);

  if (mqIn == pszString)
  {
    pszString[0] = 0;
    return;
  }

  memset (pszString, 0, iLen);

  for (i=0; i<iLen; i++)
  {
    if (mqIn[i] == ' ')
      break;

    pszString[i] = mqIn[i];
  }

  if (i == iLen)
    pszString[i] = '\0';

  return;
}

/*
 *
 */
void freeExitUserArea (PMQCXP pParms)
{
  PEXITUSERAREA   pUserExitArea  = (PEXITUSERAREA) &(pParms->ExitUserArea);
  PEXITLOCALVARS  pExitLocalVars = pUserExitArea->pExitLocalVars;
  PSECEXITDATA    pSecExitData   = pUserExitArea->pSecExitData;

  dtkDebug (pExitLocalVars, "freeExitUserArea(): Freeing allocated memory.\n");

  if (pExitLocalVars != NULL)
  {
    if (pExitLocalVars->pszTargetNameDN != NULL)
      free (pExitLocalVars->pszTargetNameDN);

    free (pExitLocalVars);
    pUserExitArea->pExitLocalVars = NULL;
  }

  if (pSecExitData != NULL)
  {
    free (pUserExitArea->pSecExitData);
    pUserExitArea->pSecExitData = NULL;
  }
}

/*
 *
 */
BOOL setINIDefaults (PEXITLOCALVARS pVars)
{
#ifdef DEBUG
  dtkDebug (NULL, "setINIDefaults(): Setting default local variables\n");
#endif
  if (!pVars)
    return 1;

  strcpy (pVars->szKeyTab, DEFAULT_KEYTAB_FILE);
  strcpy (pVars->szEntrustINIFile, DEFAULT_ENTRUST_INI_FILE);
  strcpy (pVars->szEntrustEPFFile, DEFAULT_ENTRUST_EPF_FILE);
  strcpy (pVars->szDebugFile, DEFAULT_LOG_FILE);

  pVars->cDebug = EAMS_DEBUG_NONE;
  pVars->iDebugMaxAppData = DEBUG_DEFAULT_MAX_APP_DATA;
  pVars->entSecurityMechanism = gss_mech_spkm2;
  pVars->entSymmetricKeyExchangeProtocol = SPKM_KEY_ESTB_PROTOCOL_RSA;
  pVars->bEncrypt = FALSE;
  pVars->ulReqContextLifeTime = 3600;

  return 0;
}

/*
 *
 */
char* stripLeadingAndTrailingBlanks (char* szToken)
{
  int i, iTokenLen;

  if (!szToken || !szToken [0])
    return szToken;

  iTokenLen = strlen (szToken);

  // Get rid of trailing blanks
  for (i=iTokenLen; i>0; i--)
  {
    if (szToken[i-1] == ' ' || szToken[i] == '\n')
      szToken[i-1] = 0;
    else
      break;
  }

  // Get rid of leading blanks
  for (i=0; i<iTokenLen && szToken [i] == ' '; i++)
    ;

  strcpy (szToken, szToken+i);
  return szToken;
}

/*
 *
 */
int parseLine (char* pLine, char** pszParm, char** pszValue)
{
  if (pLine[0] == '#' || pLine[0] == ';'  ||
    pLine[0] == ' ' || pLine[0] == '\r' ||
    pLine[0] == '\n')
    return 1;

  *pszParm = strtok (pLine, "=\n");

  if (*pszParm)
  {
    stripLeadingAndTrailingBlanks (*pszParm);

    *pszValue = strtok (NULL, "\r\n");
    stripLeadingAndTrailingBlanks (*pszValue);
  }

  return 0;
}

/*
 *
 */
BOOL isStringNumeric (char* pszNumber)
{
  int i, iLen;

  assert (pszNumber && *pszNumber);

  iLen = strlen (pszNumber);

  if (*pszNumber == '-' || *pszNumber == '+')
    i = 1;
  else
    i = 0;

  for (; i < iLen; i++)
  {
    if (!isdigit (pszNumber[i]))
      return FALSE;
  }

  return TRUE;
}

/*
 *  readINIFile:  Read channel INI file and set variables in PEXITLOCALVARS
 *  used in a single channel exit instance.  EXITLOCALVARS is allcated from
 *  the heap.
 */
int readINIFile (char* pszINIFile, PEXITLOCALVARS *pExitLocalVars)
{
  FILE *pF;
  char  aLine[MAX_LINE+1];
  char *pszParam, *pszValue;
  int      i, iLen;

  PEXITLOCALVARS pVars;

#ifdef DEBUG
  dtkDebug (NULL, "readINIFile(): Reading INI file\n");
#endif

  if (!pszINIFile || !pszINIFile[0])
  {
    dtkError (NULL, "SourceFile <%s> SourceLine <%d> readINIFile(): "
    "No INI file specified in Channel Definition' ExitData field.\n",
    __FILE__, __LINE__);
    return 1;
  }

  pVars = calloc (sizeof (EXITLOCALVARS), 1);
  if (pVars == NULL)
  {
    dtkError (NULL, "SourceFile <%s> SourceLine <%d> readINIFile(): "
    "Could not allocate memory for EXITLOCALVARS structure\n",
    __FILE__, __LINE__);
    return 1;
  }

  if (setINIDefaults (pVars) != 0)
  {
    dtkError (NULL, "SourceFile <%s> SourceLine <%d> readINIFile(): "
    "Could not set default values\n",
    __FILE__, __LINE__);
    return 1;
  }

  stripLeadingAndTrailingBlanks (pszINIFile);

  if ((pF = fopen (pszINIFile, "r")) == NULL)
  {
    dtkError (NULL, "SourceFile <%s> SourceLine <%d> readINIFile(): "
      "Unable to open %s.  Error (%d) <%s>\n",
      __FILE__, __LINE__, pszINIFile, errno, strerror (errno));
    return 1;
  }

  while (fgets (aLine, MAX_LINE, pF) != NULL)
  {
    if (parseLine (aLine, &pszParam, &pszValue) != 0)
      continue;
#ifdef DEBUG
    dtkDebug (NULL, "readINIFile(): Parsed line into param <%s> value <%s>\n",
      !pszParam ? "null" : pszParam, !pszValue ? "null" : pszValue);
#endif

    if (!strcmp (pszParam, "TARGET_NAME"))
    {
      if (!pszValue || pszValue[0] == 0 || strlen (pszValue) > 48)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      iLen = sizeof (pVars->mqTargetName);
      memset (pVars->mqTargetName, ' ', sizeof (pVars->mqTargetName));
      memcpy (pVars->mqTargetName, pszValue, iLen);
    }

    else if (!strcmp (pszParam, "TARGET_DN_NAME"))
    {
      if (!pszValue || pszValue[0] == 0)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      iLen = strlen (pszValue) + 1;

      if (pVars->pszTargetNameDN)
	/* If DN already allocated, free it */
	free (pVars->pszTargetNameDN);

      pVars->pszTargetNameDN = malloc (iLen);
      strcpy (pVars->pszTargetNameDN, pszValue);
    }

    else if (!strcmp (pszParam, "KEYTAB_FILE"))
    {
      if (!pszValue || pszValue[0] == 0 || strlen (pszValue) > FILENAME_MAX)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      strncpy (pVars->szKeyTab, pszValue, FILENAME_MAX);
      pVars->szKeyTab[FILENAME_MAX] = 0;
    }

    else if (!strcmp (pszParam, "ENTRUST_INI_FILE"))
    {
      if (!pszValue || pszValue[0] == 0 || strlen (pszValue) > FILENAME_MAX)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      strncpy (pVars->szEntrustINIFile, pszValue, FILENAME_MAX);
      pVars->szEntrustINIFile[FILENAME_MAX] = 0;
    }

    else if (!strcmp (pszParam, "ENTRUST_EPF_FILE"))
    {
      if (!pszValue || pszValue[0] == 0 || strlen (pszValue) > FILENAME_MAX)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      strncpy (pVars->szEntrustEPFFile, pszValue, FILENAME_MAX);
      pVars->szEntrustEPFFile[FILENAME_MAX] = 0;
    }

    else if (!strcmp (pszParam, "DEBUG_FILE"))
    {
      if (!pszValue || pszValue[0] == 0 || strlen (pszValue) > FILENAME_MAX)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      strncpy (pVars->szDebugFile, pszValue, FILENAME_MAX);
      pVars->szDebugFile[FILENAME_MAX] = 0;
    }

    else if (!strcmp (pszParam, "DEBUG_OPTIONS"))
    {
      char szTmp[MQ_EXIT_DATA_LENGTH+1], *pszToken;
      szTmp[MQ_EXIT_DATA_LENGTH] = 0;

      if (!pszValue || pszValue[0] == 0)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      strncpy (szTmp, pszValue, MQ_EXIT_DATA_LENGTH);
      pVars->cDebug = EAMS_DEBUG_NONE;

      for (i=0; i<2; i++)
      {
	pszToken = strtok ((i == 0) ? szTmp : NULL, ", ");
	if (!pszToken)
	  break;

	// If "NONE" specified turn everything off
	if (!strcmp (pszToken, EAMS_DEBUG_NONE_STRING))
	{
	  pVars->cDebug = EAMS_DEBUG_NONE;
	  break;
	}
	else if (!strcmp (pszToken, EAMS_DEBUG_FILE_STRING))
	  pVars->cDebug |= EAMS_DEBUG_FILE;
	else if (!strcmp (pszToken, EAMS_DEBUG_CONSOLE_STRING))
	  pVars->cDebug |= EAMS_DEBUG_CONSOLE;
	else
	  DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);
      }
    }

    else if (!strcmp (pszParam, "DEBUG_MAX_APP_DATA"))
    {
      if (!pszValue || pszValue[0] == 0 || strlen (pszValue) > 7)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      if (isStringNumeric (pszValue))
      {
	pVars->iDebugMaxAppData = atoi (pszValue);

	if (pVars->iDebugMaxAppData < 0 ||
	  pVars->iDebugMaxAppData > DEBUG_MAX_APP_DATA_VALUE)
	  DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	    __FILE__, __LINE__, pszValue, pszParam, pszINIFile);
      }
      else
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

    }

    else if (!strcmp (pszParam, "SECURITY_MECHANISM"))
    {
      if (!pszValue || pszValue[0] == 0)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);
      if (!strcmp (pszValue, SPKM_1_STRING))
	pVars->entSecurityMechanism = gss_mech_spkm1;
      else if (!strcmp (pszValue, SPKM_2_STRING))
	pVars->entSecurityMechanism = gss_mech_spkm2;
      else if (!strcmp (pszValue, SPKM_DEFAULT_STRING))
	pVars->entSecurityMechanism = GSS_C_NULL_OID;
      else
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);
    }

    else if (!strcmp (pszParam, "SYMMETRIC_KEY_EXCHANGE_PROTOCOL"))
    {
      if (!pszValue || pszValue[0] == 0)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      if (!strcmp (pszValue, PROTOCOL_RSA_STRING))
	pVars->entSymmetricKeyExchangeProtocol = SPKM_KEY_ESTB_PROTOCOL_RSA;
      else if (!strcmp (pszValue, PROTOCOL_D_H_STRING))
	pVars->entSymmetricKeyExchangeProtocol = SPKM_KEY_ESTB_PROTOCOL_D_H;
      else if (!strcmp (pszValue, PROTOCOL_DEFAULT_STRING))
	pVars->entSymmetricKeyExchangeProtocol = SPKM_KEY_ESTB_PROTOCOL_DEFAULT;
      else
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);
    }

    else if (!strcmp (pszParam, "SECURITY_CONTEXT_LIFE_TIME"))
    {
      if (!pszValue || pszValue[0] == 0)
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      if (isStringNumeric (pszValue))
      {
	pVars->ulReqContextLifeTime = atoi (pszValue);

	if ((int)pVars->ulReqContextLifeTime < 0)
	  DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	    __FILE__, __LINE__, pszValue, pszParam, pszINIFile);
      }
      else
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);
    }

    else if (!strcmp (pszParam, "ENCRYPTION"))
    {
      if (!pszValue || pszValue[0] == 0 ||
	(strcmp (pszValue, "Y") && strcmp (pszValue, "N")))
	DISPLAY_INI_PARM_VALUE_ERROR_AND_RETURN (
	  __FILE__, __LINE__, pszValue, pszParam, pszINIFile);

      pVars->bEncrypt = (toupper(pszValue[0]) == 'Y') ? TRUE : FALSE;
    }

    else    // Invalid parameter in INI file
    {
      dtkError (NULL, "SourceFile <%s> SourceLine <%d> readINIFile(): "
	"Found invalid parameter <%s> in INI file <%s>\n",
	__FILE__, __LINE__, pszParam, pszINIFile);
      RETURN_RC_AND_CLOSE_FILE(pF, 1);
    }
  }

#ifdef DO_WE_NEED_FOR_CLNT_CHANNEL
    // Check for required parameters that the user must supply (no defaults allowed)
  if (pVars->mqTargetName[0] == 0)
  {
    dtkError (NULL, "SourceFile <%s> SourceLine <%d> readINIFile(): "
      "Missing target name <TARGET_NAME>\n", __FILE__, __LINE__);
    RETURN_RC_AND_CLOSE_FILE(pF, 1);
  }
#endif

  if (!pVars->pszTargetNameDN || pVars->pszTargetNameDN[0] == 0)
  {
    dtkError (NULL, "SourceFile <%s> SourceLine <%d> readINIFile(): "
      "Missing target DN name <TARGET_DN_NAME>\n", __FILE__, __LINE__);
    RETURN_RC_AND_CLOSE_FILE(pF, 1);
  }
  *pExitLocalVars = pVars;

  RETURN_RC_AND_CLOSE_FILE(pF, 0);
}

/*
 *
 */
void dumpINIFields (PEXITLOCALVARS pExitLocalVars, char* pszINIFile)
{
  char szBuff[MAX_LINE+1];

  memset (szBuff, 0, MAX_LINE);

  dtkDebug (pExitLocalVars, "The following INI file <%s> fields were set:\n",     pszINIFile);
  dtkDebug (pExitLocalVars, "\tTARGET_NAME = <%s>\n",     pExitLocalVars->mqTargetName);
  dtkDebug (pExitLocalVars, "\tTARGET_DN_NAME = <%s>\n",  pExitLocalVars->pszTargetNameDN);
  dtkDebug (pExitLocalVars, "\tKEYTAB_FILE = <%s>\n",     pExitLocalVars->szKeyTab);
  dtkDebug (pExitLocalVars, "\tENTRUST_INI_FILE = <%s>\n",pExitLocalVars->szEntrustINIFile);
  dtkDebug (pExitLocalVars, "\tENTRUST_EPF_FILE = <%s>\n",pExitLocalVars->szEntrustEPFFile);
  dtkDebug (pExitLocalVars, "\tDEBUG_FILE = <%s>\n",      pExitLocalVars->szDebugFile);

  switch (pExitLocalVars->cDebug)
  {
    case EAMS_DEBUG_NONE:
      strcpy (szBuff, EAMS_DEBUG_NONE_STRING);
      break;

    case EAMS_DEBUG_FILE:
      strcpy (szBuff, EAMS_DEBUG_FILE_STRING);
      break;

    case EAMS_DEBUG_CONSOLE:
      strcpy (szBuff, EAMS_DEBUG_CONSOLE_STRING);
      break;

    case EAMS_DEBUG_FILE+EAMS_DEBUG_CONSOLE:
      strcpy (szBuff, EAMS_DEBUG_FILE_STRING);
      strcat (szBuff, ", ");
      strcat (szBuff, EAMS_DEBUG_CONSOLE_STRING);
      break;

    default:
      strcpy (szBuff, "???????");
  }

  dtkDebug (pExitLocalVars, "\tDEBUG_OPTIONS = <%s>\n", szBuff);

  if (pExitLocalVars->entSecurityMechanism == gss_mech_spkm1)
    strcpy (szBuff, SPKM_1_STRING);
  else if  (pExitLocalVars->entSecurityMechanism == gss_mech_spkm2)
    strcpy (szBuff, SPKM_2_STRING);
  else if  (pExitLocalVars->entSecurityMechanism == GSS_C_NULL_OID)
    strcpy (szBuff, SPKM_DEFAULT_STRING);
  else
    strcpy (szBuff, "???????");

  dtkDebug (pExitLocalVars, "\tDEBUG_MAX_APP_DATA = <%d> bytes\n",
      pExitLocalVars->iDebugMaxAppData);

  dtkDebug (pExitLocalVars, "\tSECURITY_MECHANISM = <%s>\n", szBuff);

  switch (pExitLocalVars->entSymmetricKeyExchangeProtocol)
  {
    case SPKM_KEY_ESTB_PROTOCOL_RSA:
      strcpy (szBuff, PROTOCOL_RSA_STRING);
      break;

    case SPKM_KEY_ESTB_PROTOCOL_D_H:
      strcpy (szBuff, PROTOCOL_D_H_STRING);
      break;

    case SPKM_KEY_ESTB_PROTOCOL_DEFAULT:
      strcpy (szBuff, PROTOCOL_DEFAULT_STRING);
      break;

    default:
      strcpy (szBuff, "???????");
  }

  dtkDebug (pExitLocalVars, "\tSYMMETRIC_KEY_EXCHANGE_PROTOCOL = <%s>\n", szBuff);

  dtkDebug (pExitLocalVars, "\tSECURITY_CONTEXT_LIFE_TIME = <%ld> seconds\n",
      pExitLocalVars->ulReqContextLifeTime);
  dtkDebug (pExitLocalVars, "\tENCRYPTION = <%s>\n", pExitLocalVars->bEncrypt ? "Y" : "N");
}

/*
 *
 */
int  getPassword (PEXITLOCALVARS pExitLocalVars,
	  char* pszKeyTab,
	  char* pszLocalQMgrName,
	  char* pszPW,
	  int   iPWLen)
{
  FILE *pF;

  assert (pszKeyTab && pszLocalQMgrName && pszPW && iPWLen);

  dtkDebug (pExitLocalVars, "getPassword(): Using keytab File <%s> for QMgr <%s>\n", pszKeyTab[0] ? pszKeyTab : "null",
      pszLocalQMgrName[0] ? pszLocalQMgrName : "null");


  if ((pF = fopen (pszKeyTab, "r")) == NULL)
  {
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> getPassword(): "
      "Unable to open %s.  Error (%d) <%s>\n",
      __FILE__, __LINE__, pszKeyTab, errno, strerror (errno));
    return 1;
  }

  if (fgets (pszPW, iPWLen, pF) == NULL)
  {
    dtkError (pExitLocalVars, "SourceFile <%s> SourceLine <%d> getPassword(): "
      "Unable to read line from file <%s>\n",
      __FILE__, __LINE__, pszKeyTab);
    RETURN_RC_AND_CLOSE_FILE(pF, 1);
  }

  pszPW = strtok (pszPW, "\r\n");
  stripLeadingAndTrailingBlanks (pszPW);

  RETURN_RC_AND_CLOSE_FILE(pF, 0);
}

/*
 *
 */
static const char* getDescription (MQLONG lCode,
	           ExitCodeEntry_t* table,
	           int iLen)
{
  int     i;
  for (i=0; i<iLen; i++)
  {
    if (lCode == table[i].lCode)
      return table[i].szDescription;
  }

  return table[--i].szDescription;
}

/*
 *
 */
const char* getExitReasonStr (MQLONG lCode)
{
  return getDescription (lCode, exitReasonTable, exitReasonTableLen);
}

/*
 *
 */
const char* getExitIdStr (MQLONG lCode)
{
  return getDescription (lCode, exitIdTable, exitIdTableLen);
}

/*
 *
 */
const char* getExitResponseStr (MQLONG lCode)
{
  return getDescription (lCode, exitResponseTable, exitResponseTableLen);
}

/*
 *
 */
static void dtkWriteDebugFile (const PEXITLOCALVARS pLocalVars, char *pszMsg)
{
  FILE *pF;
  char szErr[MAX_LINE+1];

  if ((pF = fopen (pLocalVars->szDebugFile, "a+")))
  {
    if (fputs (pszMsg, pF) < 0)
    {
      sprintf (szErr, "SourceFile <%s> SourceLine <%d> "
	"dtkWriteDebugFile(): Unexpected error from fputs() "
	"Error (%d) <%s>\n",
	__FILE__, __LINE__, pLocalVars->szDebugFile, errno, strerror (errno));
      fputs (szErr, stderr);
    }
  }
  else
  {
      sprintf (szErr, "SourceFile <%s> SourceLine <%d> "
	"dtkWriteDebugFile(): Unexpected error from fopen(\"%s\"). "
	"Error (%d) <%s>\n",
	__FILE__, __LINE__, pLocalVars->szDebugFile, errno, strerror (errno));
      fputs (szErr, stderr);
  }

  if (pF != NULL)
    fclose (pF);
}

/*
 *
 */
void dtkWriteSysLogEvent (
	    #if MQAT_DEFAULT==MQAT_WINDOWS_NT
	      WORD wEventType,
	      DWORD dwEventID,
	    #elif MQAT_DEFAULT==MQAT_UNIX
	      int iEventType, // BEN, do define for WORD ..
	    #else
	    #error Unsupported platform
	    #endif
	      char *lpszMsg)
{
  char    szSyslogBuffer[SYSLOG_BUFF_LEN+1];

#if MQAT_DEFAULT==MQAT_WINDOWS_NT
  HANDLE  hEventLog;
#endif

  strncpy (szSyslogBuffer, lpszMsg, SYSLOG_BUFF_LEN);
  szSyslogBuffer[SYSLOG_BUFF_LEN] = '\0';

  /* Get rid of newline */
  strtok (szSyslogBuffer, "\n");

#if MQAT_DEFAULT==MQAT_UNIX
  openlog (DEFAULT_IDENT, LOG_PID | LOG_CONS, LOG_AUTH);
  syslog (iEventType | LOG_AUTH, szSyslogBuffer);
  closelog ();

#elif MQAT_DEFAULT==MQAT_WINDOWS_NT
  if ((hEventLog = RegisterEventSource (NULL, DEFAULT_IDENT)))
  {
    LPSTR aszStrings[] = {szSyslogBuffer};
    int rc = ReportEvent (hEventLog,
	        wEventType,      // wEventType, i.e., info, warning, error
	        0,               // event category
	        dwEventID,       // event identifier
	        NULL,            // user security identifier (optional)
	        1,               // number of strings to merge with message
	        0,               // size of binary data, in bytes
	        (LPCTSTR*) aszStrings, // array of strings to merge with message
	        NULL);           // address of binary data
    DeregisterEventSource (hEventLog);
    if (!rc)
    {
      LPVOID lpMsgBuf;
      FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
	       NULL,
	       GetLastError(),
	       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
	       (LPTSTR) &lpMsgBuf,
	       0,
	       NULL );
      // For the time being ignore 'log full condition' error
      //MessageBox (NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION);
      LocalFree (lpMsgBuf);
    }
  }
#else
#error Unsupported platform
#endif
}


/*
 *
 */
char* prefixMsg (char*            pszOutMsg,
	 const  int       iOutMsgLen,
	 const  char* pszInMsg,
	 const  char* pszMsgPrefix)
{
  assert (pszOutMsg && iOutMsgLen > 0);

  memset (pszOutMsg, 0, iOutMsgLen);
  strncpy (pszOutMsg, pszMsgPrefix, iOutMsgLen-1);
  strncat (pszOutMsg, pszInMsg, iOutMsgLen-(strlen (pszOutMsg)+1));

  return pszOutMsg;
}

/*
 *
 */
char* getTimeStamp (char* pszOutTimeStamp)
{
  //char  aTimeStamp[26];
  char*   pszTimeStamp;

#ifdef  _POSIX_PTHREAD_SEMANTICS
  struct tm aTime;
#else
  struct tm *pTime;
#endif

  time_t aClock;

  time (&aClock);                                 /* Get time in seconds */

#ifdef  _POSIX_PTHREAD_SEMANTICS
  localtime_r (&aClock, &aTime);  /* Convert time to struct tm form */
  asctime_r (&aTime, pszOutTimeStamp);
#else
  pTime = localtime (&aClock);    /* Convert time to struct tm form */
  pszTimeStamp = asctime (pTime);
  strcpy (pszOutTimeStamp, pszTimeStamp);
#endif

  return pszOutTimeStamp;
}

/*
 *
 */
static void dtkWrite (const PEXITLOCALVARS pLocalVars,
	    enum  MsgType_t msgType,
	    char* lpszMsg)
{
  char    szMsg[MAX_LINE+1];
  char*   pszPrefix;
#if MQAT_DEFAULT==MQAT_WINDOWS_NT
  WORD    wEventType;
  DWORD   dwEventID;
#else // UNIX
  int             wEventType;
#endif

  //BOOL  bSysLog = FALSE;

  char    aTimeStamp[TIMESTAMP_LEN];

  /* Declare just incase other message types will require timesampe */
  //char  szTSPrefix[TIMESTAMP_LEN + sizeof (MQSECCE_DEBUG_MSG_PREFIX)];

  switch (msgType)
  {
    case DTK_INFO:
      pszPrefix       = MQSECCE_INFO_MSG_PREFIX;

#if MQAT_DEFAULT==MQAT_WINDOWS_NT
      wEventType      = EVENTLOG_INFORMATION_TYPE;
      dwEventID       = MSG_INFO;
#else   // UNIX
      wEventType      = LOG_INFO;
#endif
      break;

    case DTK_WARNING:
      pszPrefix       = MQSECCE_WARNING_MSG_PREFIX;

#if MQAT_DEFAULT==MQAT_WINDOWS_NT
      wEventType      = EVENTLOG_WARNING_TYPE;
      dwEventID       = MSG_WARNING;
#else   // UNIX
      wEventType      = LOG_WARNING;
#endif

      break;

    case DTK_ERROR:
      pszPrefix = MQSECCE_ERROR_MSG_PREFIX;


#if MQAT_DEFAULT==MQAT_WINDOWS_NT
      wEventType      = EVENTLOG_ERROR_TYPE;
      dwEventID       = MSG_ERROR;
#else   // UNIX
      wEventType      = LOG_ERR;
#endif

      break;

    case DTK_DEBUG:
      //pszPrefix = MQSECCE_DEBUG_MSG_PREFIX;
      getTimeStamp (aTimeStamp);
      aTimeStamp[TIMESTAMP_LEN-2] = ' '; // replace '\n' with ' '
      pszPrefix = aTimeStamp;
      wEventType = (WORD) -1; // 0xffff
      break;

    case DTK_DEBUG_NO_PREFIX:
      pszPrefix = "";
      wEventType = (WORD) -1; // 0xffff
      break;

    default:
      pszPrefix = "(unknown prefix) ";
      wEventType = (WORD) -1; // 0xffff
  }

  prefixMsg (szMsg, MAX_LINE, lpszMsg, pszPrefix);

  if (pLocalVars)
  {

    if (pLocalVars->cDebug & (EAMS_DEBUG_FILE | EAMS_DEBUG_CONSOLE))
    {
      if (pLocalVars->cDebug & EAMS_DEBUG_FILE)
	dtkWriteDebugFile (pLocalVars, szMsg);

      if (pLocalVars->cDebug & EAMS_DEBUG_CONSOLE)
	fputs (szMsg, stdout);
    }
  }
  else
  { // Not fully initialized --- therefore, can not write to debug/log file
    // If not to be sent to syslog/eventlog, just try console.
#ifdef DEBUG
    printf (szMsg);
#else
    ; /* for total silence */
#endif
  }

  // If syslog/eventlog is desired do it.
  if (wEventType != (WORD) -1)
#if MQAT_DEFAULT==MQAT_WINDOWS_NT
    dtkWriteSysLogEvent (wEventType, dwEventID, szMsg);
#else
    dtkWriteSysLogEvent (wEventType, szMsg);
#endif

}

/*
 *
 */
void dtkDisplay (const PEXITLOCALVARS pLocalVars, char *fmt, ...)
{
  char szMsg[MAX_LINE+1];
  va_list ap;
  va_start (ap, fmt);

  vsprintf (szMsg, fmt, ap);
  dtkWrite (pLocalVars, DTK_INFO, szMsg);

  va_end (ap);
}

/*
 *
 */
void dtkWarning (const PEXITLOCALVARS pLocalVars, char *fmt, ...)
{
  char szMsg[MAX_LINE+1];
  va_list ap;
  va_start (ap, fmt);

  vsprintf (szMsg, fmt, ap);
  dtkWrite (pLocalVars, DTK_WARNING, szMsg);

  va_end (ap);
}

/*
 *
 */
void dtkError (const PEXITLOCALVARS pLocalVars, char *fmt, ...)
{
  char szMsg[MAX_LINE+1];
  va_list ap;
  va_start (ap, fmt);

  vsprintf (szMsg, fmt, ap);
  dtkWrite (pLocalVars, DTK_ERROR, szMsg);

  va_end (ap);
}

/*
 *
 */
void dtkDebug (const PEXITLOCALVARS pLocalVars, char *fmt, ...)
{
  char szMsg[MAX_LINE+1];
  va_list ap;
  va_start (ap, fmt);

  vsprintf (szMsg, fmt, ap);
  dtkWrite (pLocalVars, DTK_DEBUG, szMsg);

  va_end (ap);
}

/*
 *      Display debug message without prefix
 */
void dtkDebug2 (const PEXITLOCALVARS pLocalVars, char *fmt, ...)
{
  char szMsg[MAX_LINE+1];
  va_list ap;
  va_start (ap, fmt);

  vsprintf (szMsg, fmt, ap);
  dtkWrite (pLocalVars, DTK_DEBUG_NO_PREFIX, szMsg);

  va_end (ap);
}
